"""
The ``mdopt.mps.utils`` module.
===========================================
This module contains various MPS utilities.
"""

from typing import Union, Optional
from functools import reduce
import numpy as np
from opt_einsum import contract
from mdopt.utils.utils import svd
from mdopt.mps.canonical import CanonicalMPS
from mdopt.mps.explicit import ExplicitMPS


def create_state_vector(num_sites: np.int16, phys_dim: np.int16 = 2) -> np.ndarray:
    """
    Creates a uniform random complex quantum state in the form of a state vector.

    Parameters
    ----------
    num_sites : np.int16
        Number of degrees of freedom.
    phys_dim : np.int16
        Number of dimensions of each degree of freedom.

    Returns
    -------
    state_vector : np.ndarray
        The resulting state vector.
    """

    state_vector = np.random.uniform(
        size=(phys_dim**num_sites)
    ) + 1j * np.random.uniform(size=(phys_dim**num_sites))
    state_vector /= np.linalg.norm(state_vector)

    return state_vector


def find_orth_centre(
    mps: CanonicalMPS, return_orth_flags: bool = False, tolerance: np.float64 = 1e-12
):
    """
    Returns a list of integers corresponding to positions of orthogonality centres of an MPS.

    Parameters
    ----------
    mps : CanonicalMPS
        The MPS to find the orthogonality centre(s) in.
    return_orth_flags : bool
        Whether to return if each tensor is a right or a left isometry.
    tolerance : np.float64
        Numerical tolerance for checking the isometry property.

    Returns
    -------
    orth_centres : list[np.int16]
        Indices of sites at which tensors are orthogonality centres.
    orth_flags_left : Optional[list[bool]]
        Boolean variables for each tensor corresponding to being a left isometry.
    orth_flags_right : Optional[list[bool]]
        Boolean variables for each tensor corresponding to being a right isometry.

    Raises
    ------
    ValueError
        If an :class:`ExplicitMPS` instance is passed as an input.
        They do not have orthogonality centres by definition.
    """

    if isinstance(mps, ExplicitMPS):
        raise ValueError(
            "Orthogonality centre is undefined for an Explicit MPS instance."
        )

    num_sites = len(mps)

    orth_flags_left = []
    orth_flags_right = []
    orth_centres = []

    for i, tensor in enumerate(mps.tensors):

        to_be_identity_left = contract(
            "ijk, ijl -> kl", tensor, np.conjugate(tensor), optimize=[(0, 1)]
        )
        to_be_identity_right = contract(
            "ijk, ljk -> il", tensor, np.conjugate(tensor), optimize=[(0, 1)]
        )

        identity_left = np.identity(to_be_identity_left.shape[0])
        identity_right = np.identity(to_be_identity_right.shape[0])

        norm_left = np.linalg.norm(to_be_identity_left - identity_left)
        norm_right = np.linalg.norm(to_be_identity_right - identity_right)

        orth_flags_left.append(np.isclose(norm_left, 0, atol=tolerance))
        orth_flags_right.append(np.isclose(norm_right, 0, atol=tolerance))

        if not (np.isclose(norm_left, 0, atol=tolerance)) and not (
            np.isclose(norm_right, 0, atol=tolerance)
        ):
            orth_centres.append(i)

    # Handling exceptions, right- and left-canonical forms, and cases
    # when the orthogonality centre might be left- or right- isometry at
    # the boundaries while all the other tensors are the opposite isometries.
    if orth_flags_left in ([True] + [False] * (num_sites - 1), [False] * num_sites):
        if orth_flags_right == [not flag for flag in orth_flags_left]:
            orth_centres.append(0)

    if orth_flags_left in ([True] * (num_sites - 1) + [False], [True] * num_sites):
        if orth_flags_right == [not flag for flag in orth_flags_left]:
            orth_centres.append(num_sites - 1)

    # Handling a product state.
    if (orth_flags_left == [True] * num_sites) and (
        orth_flags_right == [True] * num_sites
    ):
        orth_centres.append(0)

    if return_orth_flags:
        return orth_centres, orth_flags_left, orth_flags_right

    return orth_centres


def is_canonical(mps: CanonicalMPS, tolerance: np.float64 = 1e-12):
    """
    Checks if the MPS is in any of the canonical forms.

    Parameters
    ----------
    mps : CanonicalMPS
        The MPS to check the canonical form of.
    tolerance : np.float64
        Numerical tolerance for checking the isometry property.

    Returns
    -------
    if_canonical : bool
        ``True`` if the MPS is in any of the canonical forms.

    Raises
    ------
    ValueError
        If an :class:`ExplicitMPS` instance is passed as an input.
        They do not have orthogonality centres by definition.
    """

    if isinstance(mps, ExplicitMPS):
        raise ValueError(
            "Orthogonality centre is undefined for an Explicit MPS instance."
        )

    flags_left = []
    flags_right = []
    for _, tensor in enumerate(mps.tensors):

        to_be_identity_left = contract(
            "ijk, ijl -> kl", tensor, np.conjugate(tensor), optimize=[(0, 1)]
        )
        to_be_identity_right = contract(
            "ijk, ljk -> il", tensor, np.conjugate(tensor), optimize=[(0, 1)]
        )

        identity_left = np.identity(to_be_identity_left.shape[0], dtype=np.float64)
        identity_right = np.identity(to_be_identity_right.shape[0], dtype=np.float64)

        flags_left.append(
            np.isclose(
                np.linalg.norm(to_be_identity_left - identity_left), 0, atol=tolerance
            )
        )
        flags_right.append(
            np.isclose(
                np.linalg.norm(to_be_identity_right - identity_right), 0, atol=tolerance
            )
        )

    if np.array(flags_left).all() or np.array(flags_right).all():
        return True

    orth_centres = find_orth_centre(mps)
    if_canonical = len(orth_centres) == 1

    return if_canonical


def inner_product(
    mps_1: Union[CanonicalMPS, ExplicitMPS], mps_2: Union[CanonicalMPS, ExplicitMPS]
) -> Union[np.float64, np.complex128]:
    """
    Returns an inner product between 2 Matrix Product States.

    Parameters
    ----------
    mps_1 : Union[CanonicalMPS, ExplicitMPS]
        The first MPS in the inner product.
    mps_1 : Union[CanonicalMPS, ExplicitMPS]
        The second MPS in the inner product.

    Returns
    -------
    product : Union[np.float64, np.complex128]
        The value of the inner product.

    Raises
    ------
    ValueError
        If the Matrix Product States are of different length.
    """

    if len(mps_1) != len(mps_2):
        raise ValueError(
            f"The number of sites in the first MPS is {len(mps_1)} while "
            f"the number of sites in the second MPS is {len(mps_2)}. "
            "The MPS's must be of equal length."
        )

    num_sites = len(mps_1)

    mps_1 = mps_1.conjugate()

    if isinstance(mps_1, ExplicitMPS):
        mps_1 = mps_1.right_canonical()
    if isinstance(mps_2, ExplicitMPS):
        mps_2 = mps_2.right_canonical()

    tensors = []

    for i in range(num_sites):

        dims_1 = mps_1.tensors[i].shape
        dims_2 = mps_2.tensors[i].shape

        tensors.append(
            contract(
                "ijk, ljm -> ilkm",
                mps_1.tensors[i],
                mps_2.tensors[i],
                optimize=[(0, 1)],
            ).reshape((dims_1[0] * dims_2[0], dims_1[2] * dims_2[2]))
        )

    product = reduce(lambda a, b: np.tensordot(a, b, (-1, 0)), tensors)

    return product[0][0]


def mps_from_dense(
    state_vector: np.ndarray,
    phys_dim: np.int16 = 2,
    chi_max: np.int16 = 1e4,
    tolerance: np.float64 = 1e-12,
    form: str = "Explicit",
    orth_centre: Optional[np.int16] = None,
) -> Union[CanonicalMPS, ExplicitMPS]:
    """
    Builds an MPS from a dense (state-vector) from.

    Parameters
    ----------
    state_vector : np.ndarray
        The initial state vector.
    phys_dim : np.int16
        Dimensionality of the local Hilbert space, i.e.,
        the dimension of each physical leg of the MPS.
    chi_max : np.int16
        Maximum number of singular values to keep.
    tolerance : np.float64
        Absolute tolerance of the normalisation of the singular value spectrum at each bond.
    form : str
        The form of the MPS. Available options:
            | ``Explicit`` : The :class:`ExplicitMPS` form (by default).
            | ``Right-canonical`` : The :class:`CanonicalMPS` right-canonical form.
            | ``Left-canonical`` : The :class:`CanonicalMPS` left-canonical form.
            | ``Mixed-canonical`` : The :class:`CanonicalMPS` mixed-canonical form.
    orth_centre : Optional[np.int16]
        The orthogonal centre position for the mixed-canonical form.

    Returns
    -------
    mps : Union[CanonicalMPS, ExplicitMPS]
        The resulting MPS.

    Raises
    ------
    ValueError
        If the vector's length does not correspond to the physical dimension.
    """

    state_vector = np.copy(state_vector)

    if state_vector.flatten().shape[0] % phys_dim != 0:
        raise ValueError(
            "The dimension of the flattened state vector is incorrect "
            "(does not correspond to the product of local dimensions)."
        )

    tensors = []
    singular_values = []

    state_vector = state_vector.reshape((-1, phys_dim))

    state_vector, singular_values_local, v_r = svd(
        state_vector, chi_max=chi_max, renormalise=False
    )

    tensors.append(np.expand_dims(v_r, -1))
    singular_values.append(singular_values_local)

    while state_vector.shape[0] >= phys_dim:

        state_vector = np.matmul(state_vector, np.diag(singular_values_local))

        bond_dim = state_vector.shape[-1]
        state_vector = state_vector.reshape((-1, phys_dim * bond_dim))
        state_vector, singular_values_local, v_r = svd(
            state_vector, chi_max=chi_max, renormalise=False
        )
        v_r = v_r.reshape((-1, phys_dim, bond_dim))

        tensors.insert(0, v_r)
        singular_values.insert(0, singular_values_local)

    singular_values.append(np.array([1.0]))

    for i, _ in enumerate(tensors):

        tensors[i] = np.tensordot(
            tensors[i], np.linalg.inv(np.diag(singular_values[i + 1])), (2, 0)
        )

    if form == "Right-canonical":
        return ExplicitMPS(
            tensors, singular_values, tolerance=tolerance, chi_max=chi_max
        ).right_canonical()
    if form == "Left-canonical":
        return ExplicitMPS(
            tensors, singular_values, tolerance=tolerance, chi_max=chi_max
        ).left_canonical()
    if form == "Mixed-canonical":
        return ExplicitMPS(
            tensors, singular_values, tolerance=tolerance, chi_max=chi_max
        ).mixed_canonical(orth_centre=orth_centre)

    return ExplicitMPS(tensors, singular_values, tolerance=tolerance)


def create_simple_product_state(
    num_sites: np.int16,
    which: str = "0",
    phys_dim: np.int16 = 2,
    form: str = "Explicit",
) -> Union[CanonicalMPS, ExplicitMPS]:
    r"""
    Creates a simple product-state MPS.

    Parameters
    ----------
    num_sites : np.int16
        The number of sites.
    which : str
        The form of the MPS, for explanation see the notes. Available options:
            | ``0`` : The :math:`|0...0>` state.
            | ``1`` : The :math:`|1...1>` state.
            | ``+`` : The :math:`|+...+>` state.
    phys_dim : np.int16
        Dimensionality of the local Hilbert space, i.e.,
        the dimension of each physical leg of the MPS.
    form : str
        The form of the MPS. Available options:
            | ``Explicit`` : The :class:`ExplicitMPS` form (by default).
            | ``Right-canonical`` : The :class:`CanonicalMPS` right-canonical form.
            | ``Left-canonical`` : The :class:`CanonicalMPS` left-canonical form.

    Returns
    -------
    mps : Union[CanonicalMPS, ExplicitMPS]
        The resulting MPS.

    Raises
    ------
    ValueError
        If the chosen form is mixed-canonical.
        This form is not available for product states.

    Notes
    -----
    Produces a Matrix Product State consisting of tensors with bond dimenstions equal to 1.
    The tensors are defined as follows:
        | :math:`| 0 \rangle = \underbrace{(1, 0, ..., 0, 0)}_{\text{phys_dim}}`,
        | :math:`| 1 \rangle = \underbrace{(0, 0, ..., 0, 1)}_{\text{phys_dim}}`,
        | :math:`| + \rangle = \underbrace{(\frac{1}{\sqrt{\text{phys_dim}}}, ..., \frac{1}{\sqrt{\text{phys_dim}}})}_{\text{phys_dim}}`.
    """

    tensor = np.zeros((phys_dim,))
    if which == "0":
        tensor[0] = 1.0
    if which == "1":
        tensor[-1] = 1.0
    if which == "+":
        for i in range(phys_dim):
            tensor[i] = 1 / np.sqrt(phys_dim)

    tensors = [tensor.reshape((1, phys_dim, 1)) for _ in range(num_sites)]
    singular_values = [[1.0] for _ in range(num_sites + 1)]

    if form == "Right-canonical":
        return ExplicitMPS(tensors, singular_values).right_canonical()
    if form == "Left-canonical":
        return ExplicitMPS(tensors, singular_values).left_canonical()
    if form == "Mixed-canonical":
        raise ValueError("Mixed-canonical form is not defined for a product state.")

    return ExplicitMPS(tensors, singular_values)


def create_custom_product_state(
    string: str, phys_dim: np.int16 = 2, form: str = "Explicit"
) -> Union[CanonicalMPS, ExplicitMPS]:
    r"""
    Creates a custom product-state MPS defined by the ``string`` argument.

    Parameters
    ----------
    string : str
        The string defining the product-state MPS. Available characters: ``0``, ``1``, ``+``.
    phys_dim : np.int16
        Dimensionality of the local Hilbert space, i.e.,
        the dimension of each physical leg of the MPS.
    form : str
        The form of the MPS. Available options:
            | ``Explicit`` : The :class:`ExplicitMPS` form (by default).
            | ``Right-canonical`` : The :class:`CanonicalMPS` right-canonical form.
            | ``Left-canonical`` : The :class:`CanonicalMPS` left-canonical form.

    Returns
    -------
    mps : Union[CanonicalMPS, ExplicitMPS]
        The resulting MPS.

    Raises
    ------
    ValueError
        If a symbol inside the ``string`` argument does not belong to the available set.
    ValueError
        If the chosen form is mixed-canonical.
        This form is not available for product states.

    Notes
    -----
    Produces a Matrix Product State consisting of tensors with bond dimenstions equal to 1.
    The tensors are defined as follows:
        | :math:`| 0 \rangle = \underbrace{(1, 0, ..., 0, 0)}_{\text{phys_dim}}`,
        | :math:`| 1 \rangle = \underbrace{(0, 0, ..., 0, 1)}_{\text{phys_dim}}`,
        | :math:`| + \rangle = \underbrace{(\frac{1}{\sqrt{\text{phys_dim}}}, ..., \frac{1}{\sqrt{\text{phys_dim}}})}_{\text{phys_dim}}`.
    The state is renormalized at the end.
    """

    num_sites = len(string)
    tensors = []

    for symbol in string:

        tensor = np.zeros((phys_dim,))
        if symbol == "0":
            tensor[0] = 1.0
        elif symbol == "1":
            tensor[-1] = 1.0
        elif symbol == "+":
            for i in range(phys_dim):
                tensor[i] = 1 / np.sqrt(phys_dim)
        else:
            raise ValueError(
                f"The string argument accepts only 0, 1 or + as elements, given {symbol}."
            )
        tensors.append(tensor)

    tensors = [tensor.reshape((1, phys_dim, 1)) for tensor in tensors]
    tensors /= np.linalg.norm(tensors)
    singular_values = [[1.0] for _ in range(num_sites + 1)]

    if form == "Right-canonical":
        return ExplicitMPS(tensors, singular_values).right_canonical()
    if form == "Left-canonical":
        return ExplicitMPS(tensors, singular_values).left_canonical()
    if form == "Mixed-canonical":
        raise ValueError("Mixed-canonical form is not defined for a product state.")

    return ExplicitMPS(tensors, singular_values)



"""
This module contains the :class:`DephasingDMRG` and the :class:`EffectiveDensityOperator` classes.

This algorithm's main feature is that it restricts the target-state search to
the computational basis states domain.
In particular, we use it to find the main component of a Matrix Density Product Operator (MDPO),
i.e., a computational basis state contributing the largest amplitude.

In our notation, MDPO for ``n`` sites denotes the following object::

         |      |               |       |
         |      |               |       |
    ----(0*)---(1*)--- ... ---(n-2*)--(n-1*)---
    ----(0)----(1)---- ... ---(n-2)---(n-1)----
         |      |               |       |
         |      |               |       |

An MDPO is formed by an MPS and its complex-conjugated version.
The main idea is to find the main component of this object without
performing the kronecker product explicitly.
"""


from typing import Union
import numpy as np
import scipy.sparse
from opt_einsum import contract
from scipy.sparse.linalg import eigsh
from tqdm import tqdm

from mdopt.mps.canonical import CanonicalMPS
from mdopt.mps.explicit import ExplicitMPS
from mdopt.utils.utils import split_two_site_tensor


class EffectiveDensityOperator(scipy.sparse.linalg.LinearOperator):
    """
    Class to store an effective two-site density operator.

    To take more advantage of the ``scipy.sparse.linalg`` module, we make a special class
    for local effective density operators extending the analogy from local effective operators.
    It allows us to compute eigenvectors more effeciently.

    The diagram displaying the contraction can be found in the supplementary notes.
    """

    def __init__(
        self,
        left_environment: np.ndarray,
        mps_target_1: np.ndarray,
        mps_target_2: np.ndarray,
        right_environment: np.ndarray,
    ):
        """
        Initialise an effective dephased density operator tensor network.

        Parameters
        ----------
        left_environment : np.ndarray
            The left environment for the effective dephased density operator.
        mps_target_1 : np.ndarray
            The left target matrix product state tensor.
        mps_target_2 : np.ndarray
            The right target matrix product state tensor.
        right_environment : np.ndarray
            The right environment for the effective dephased density operator.

        Raises
        ------
        ValueError
            If the left environment tensor is not four-dimensional.
        ValueError
            If the first target MPS tensor is not three-dimensional.
        ValueError
            If the second target MPS tensor is not three-dimensional.
        ValueError
            If the right environment tensor is not four-dimensional.
        """
        if len(left_environment.shape) != 4:
            raise ValueError(
                "A valid left environment tensor must have 4 legs"
                f"while the one given has {len(left_environment.shape)}."
            )
        if len(mps_target_1.shape) != 3:
            raise ValueError(
                "A valid target MPS tensor must have 3 legs"
                f"while the one given has {len(mps_target_1.shape)}."
            )
        if len(mps_target_2.shape) != 3:
            raise ValueError(
                "A valid target MPS tensor must have 3 legs"
                f"while the one given has {len(mps_target_1.shape)}."
            )
        if len(right_environment.shape) != 4:
            raise ValueError(
                "A valid right environment tensor must have 4 legs"
                f"while the one given has {len(right_environment.shape)}."
            )

        self.left_environment = left_environment
        self.right_environment = right_environment
        self.mps_target_1 = mps_target_1
        self.mps_target_2 = mps_target_2
        chi_1, chi_2 = (
            left_environment.shape[3],
            right_environment.shape[3],
        )
        d_1, d_2 = (
            mps_target_1.shape[1],
            mps_target_2.shape[1],
        )
        self.x_shape = (chi_1, d_1, d_2, chi_2)
        self.shape = (chi_1 * d_1 * d_2 * chi_2, chi_1 * d_1 * d_2 * chi_2)
        self.dtype = mps_target_1.dtype
        super().__init__(shape=self.shape, dtype=self.dtype)

    def _matvec(self, x: np.ndarray) -> np.ndarray:
        """
        Performs matrix-vector multiplication.

        Computes ``effective_density_operator * |x> = |x'>``.
        This function is used by :func:`scipy.sparse.linalg.eigsh` to diagonalise
        the effective density operator with the Lanczos method, withouth generating the full matrix.

        Parameters
        ----------
        x : np.ndarray
            The two-site tensor to be acted on by an effective density operator.
        """

        two_site_tensor = np.reshape(x, self.x_shape)

        if len(two_site_tensor.shape) != 4:
            raise ValueError(
                f"A valid two-site tensor must have 4 legs"
                f"while the one given has {len(two_site_tensor.shape)}."
            )

        copy_tensor = np.fromfunction(
            lambda i, j, k: np.logical_and(i == j, j == k), (2, 2, 2), dtype=np.int16
        )

        einsum_string = (
            "ustw, ailu, ifj, jhk, bef, cgh, lom, mpn, eso, gtp, dknw -> abcd"
        )
        two_site_tensor = contract(
            einsum_string,
            two_site_tensor,
            self.left_environment,
            np.conjugate(self.mps_target_1),
            np.conjugate(self.mps_target_2),
            copy_tensor,
            copy_tensor,
            self.mps_target_1,
            self.mps_target_2,
            copy_tensor,
            copy_tensor,
            self.right_environment,
            optimize=[
                (0, 8),
                (0, 1),
                (0, 6),
                (0, 5),
                (1, 2),
                (2, 3),
                (3, 4),
                (2, 3),
                (1, 2),
                (0, 1),
            ],
        )

        return np.reshape(two_site_tensor, self.shape[0])


class DephasingDMRG:
    """
    Class holding the Dephasing Density Matrix Renormalisation Group algorithm with two-site updates
    for a finite-size system with open-boundary conditions.

    Attributes
    ----------
    mps : Union[ExplicitMPS, CanonicalMPS]
        MPS serving as a current approximation of the target state.
    mps_target : Union[ExplicitMPS, CanonicalMPS]
        The target MPS in the right-canonical form.
        This MPS is used to construct the dephased MDPO.
    chi_max : np.int16
        The highest bond dimension of an MPS allowed.
    mode : str
        The eigensolver mode. Available options:
            | ``LM`` : Largest (in magnitude) eigenvalues.
            | ``SM`` : Smallest (in magnitude) eigenvalues.
            | ``LA`` : Largest (algebraic) eigenvalues.
            | ``SA`` : Smallest (algebraic) eigenvalues.
    cut : np.float64
        The lower boundary of the spectrum, i.e., all
        the singular values smaller than that will be discarded.
    silent : bool
        Whether to show/hide the progress bar.
    """

    def __init__(
        self,
        mps: Union[ExplicitMPS, CanonicalMPS],
        mps_target: Union[ExplicitMPS, CanonicalMPS],
        chi_max: np.int16 = 1e4,
        cut: np.float64 = 1e-12,
        mode: str = "SA",
        silent: bool = False,
        copy: bool = True,
    ):
        if len(mps) != len(mps_target):
            raise ValueError(
                f"The MPS has length {len(mps)},"
                f"the target MPS has length {len(mps_target)},"
                "but the lengths should be equal."
            )
        if copy:
            self.mps = mps.copy()
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.right_canonical()
        self.mps = mps
        self.left_environments = [None] * len(mps)
        self.right_environments = [None] * len(mps)
        mps_target = mps_target.right_canonical()
        self.mps_target = mps_target
        self.chi_max = chi_max
        self.cut = cut
        self.mode = mode
        self.silent = silent

        start_bond_dim = self.mps_target.tensors[0].shape[0]
        chi = mps.tensors[0].shape[0]
        left_environment = np.zeros(
            [chi, start_bond_dim, start_bond_dim, chi], dtype=np.float64
        )
        right_environment = np.zeros(
            [chi, start_bond_dim, start_bond_dim, chi], dtype=np.float64
        )
        left_environment[:, 0, 0, :] = np.eye(chi, dtype=np.float64)
        right_environment[:, start_bond_dim - 1, start_bond_dim - 1, :] = np.eye(
            chi, dtype=np.float64
        )
        self.left_environments[0] = left_environment
        self.right_environments[-1] = right_environment
        for i in reversed(range(1, len(mps))):
            self.update_right_environment(i)

    def sweep(self):
        """
        One Dephasing DMRG sweep.

        A method performing one Dephasing DMRG sweep, which consists of
        two series of ``update_bond`` sweeps which go back and forth.
        """

        for i in range(self.mps.num_sites - 1):
            self.update_bond(i)

        for i in reversed(range(self.mps.num_sites - 1)):
            self.update_bond(i)

    def update_bond(self, i: np.int16):
        """
        Updates the bond between sites ``i`` and ``i+1``.
        """

        j = i + 1

        effective_density_operator = EffectiveDensityOperator(
            self.left_environments[i],
            self.mps_target.tensors[i],
            self.mps_target.tensors[j],
            self.right_environments[j],
        )

        if isinstance(self.mps, ExplicitMPS):
            initial_guess = self.mps.two_site_right_iso(i).reshape(
                effective_density_operator.shape[0]
            )
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i)
            initial_guess = self.mps.two_site_tensor_next(i).reshape(
                effective_density_operator.shape[0]
            )

        _, eigenvectors = eigsh(
            effective_density_operator,
            k=1,
            which=self.mode,
            return_eigenvectors=True,
            v0=initial_guess,
            tol=1e-8,
        )
        x = eigenvectors[:, 0].reshape(effective_density_operator.x_shape)
        left_iso_i, singular_values_j, right_iso_j = split_two_site_tensor(
            x, chi_max=self.chi_max, cut=self.cut, renormalise=True
        )

        if isinstance(self.mps, CanonicalMPS):
            self.mps.tensors[i] = np.tensordot(
                left_iso_i, np.diag(singular_values_j), (1, 0)
            )
            self.mps.orth_centre = i
            self.mps.tensors[j] = right_iso_j

        if isinstance(self.mps, ExplicitMPS):
            self.mps.tensors[i] = np.tensordot(
                np.linalg.inv(np.diag(self.mps.singular_values[i])), left_iso_i, (1, 0)
            )
            self.mps.tensors[j] = np.tensordot(
                right_iso_j,
                np.linalg.inv(np.diag(self.mps.singular_values[j + 1])),
                (2, 0),
            )
            self.mps.singular_values[j] = singular_values_j

        self.update_left_environment(i)
        self.update_right_environment(j)

    def update_right_environment(self, i: np.int16):
        """
        Compute the ``right_environment`` right of site ``i-1``
        from the ``right_environment`` right of site ``i``.
        """

        right_environment = self.right_environments[i]

        if isinstance(self.mps, ExplicitMPS):
            right_iso = self.mps.single_site_right_iso(i)
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i - 1)
            right_iso = self.mps.single_site_tensor(i)

        right_environment = contract(
            "ijkl, omi, pmj, qnk, rnl -> opqr",
            right_environment,
            right_iso,
            np.conjugate(self.mps_target.tensors[i]),
            self.mps_target.tensors[i],
            np.conjugate(right_iso),
            optimize=[(0, 2), (0, 1), (0, 1), (0, 1)],
        )
        self.right_environments[i - 1] = right_environment

    def update_left_environment(self, i: np.int16):
        """
        Compute the ``left_environment`` left of site ``i+1``
        from the  ``left_environment`` left of site ``i``.
        """

        left_environment = self.left_environments[i]

        if isinstance(self.mps, ExplicitMPS):
            left_iso = self.mps.single_site_left_iso(i)
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i + 1)
            left_iso = self.mps.single_site_tensor(i)

        left_environment = contract(
            "ijkl, imo, jmp, knq, lnr -> opqr",
            left_environment,
            left_iso,
            np.conjugate(self.mps_target.tensors[i]),
            self.mps_target.tensors[i],
            np.conjugate(left_iso),
            optimize=[(0, 2), (0, 1), (0, 1), (0, 1)],
        )
        self.left_environments[i + 1] = left_environment

    def run(self, num_iter: np.int16 = 1):
        """
        Run the algorithm, i.e., run the ``sweep`` method for ``num_iter`` number of times.
        """

        for _ in tqdm(range(num_iter), disable=self.silent):
            self.sweep()



"""This module contains miscellaneous utilities."""

import numpy as np
import scipy
from opt_einsum import contract


def svd(
    mat: np.ndarray,
    cut: np.float64 = 1e-12,
    chi_max: np.int16 = 1e4,
    renormalise: bool = False,
) -> tuple[np.ndarray]:
    """
    Performs the Singular Value Decomposition with different features.

    Parameters
    ----------
    mat : np.ndarray
        Matrix provided as a tensor with 2 dimensions.
    cut : np.float64
        Singular values smaller than this will be discarded.
    chi_max : np.int16
        Maximum number of singular values to keep.
    renormalise : bool
        Whether to renormalise the singular value spectrum after the cut.

    Returns
    -------
    u_l : np.ndarray
        Unitary matrix having left singular vectors as columns.
    singular_values : np.ndarray
        The singular values, sorted in non-increasing order.
    v_r : np.ndarray
        Unitary matrix having right singular vectors as rows.

    Raises
    ------
    ValueError
        If the `np.ndarray` provided is not two-dimensional.
    """

    if len(mat.shape) != 2:
        raise ValueError(
            f"A valid matrix must have 2 dimensions while the one given has {len(mat.shape)}."
        )
    try:
        u_l, singular_values, v_r = np.linalg.svd(
            mat, full_matrices=False, compute_uv=True, hermitian=False
        )
    except np.linalg.LinAlgError:
        u_l, singular_values, v_r = scipy.linalg.svd(
            mat, full_matrices=False, compute_uv=True, lapack_driver="gesvd"
        )

    max_num = min(chi_max, np.sum(singular_values > cut))
    ind = np.argsort(singular_values)[::-1][:max_num]

    u_l, singular_values, v_r = u_l[:, ind], singular_values[ind], v_r[ind, :]

    if renormalise:
        singular_values /= np.linalg.norm(singular_values)

    return u_l, singular_values, v_r


def kron_tensors(
    tensor_1: np.ndarray,
    tensor_2: np.ndarray,
    conjugate_second: bool = False,
    merge_physicals: bool = True,
) -> np.ndarray:
    """
    Computes a kronecker product of 2 MPS tensors.

    Parameters
    ----------
    tensor_1 : np.ndarray
        The first tensor of the product.
    tensor_2 : np.ndarray
        The second tensor of the product.
    conjugate_second : bool
        Whether to complex-conjugate the second tensor.
    merge_physicals : bool
        Whether to merge physical indices.

    Returns
    -------
    product
        The resulting kronecker product.

    Raises
    ------
    ValueError
        If the first MPS tensor is not three-dimensional.
    ValueError
        If the second MPS tensor is not three-dimensional.

    Notes
    -----
    This function acts according to the following diagram::

           tensor_2
              i
              |                      i
        j ---( )--- k                |
                        -->    jl---( )---kn
        l ---( )--- n                |
              |                      m
              m
           tensor_1

    The legs of the resulting tensor are indexed as ``(jl, m, i, kn)``.
    Indices `i` and `m` can be merged if `merge_physicals=True`.
    """

    if len(tensor_1.shape) != 3:
        raise ValueError(
            f"The number of dimensions of the first tensor is {len(tensor_1.shape)},"
            "but the number of dimensions expected is 3."
        )
    if len(tensor_2.shape) != 3:
        raise ValueError(
            f"The number of dimensions of the second tensor is {len(tensor_2.shape)},"
            "but the number of dimensions expected is 3."
        )

    if conjugate_second:
        product = np.kron(tensor_1, np.conjugate(tensor_2))
    else:
        product = np.kron(tensor_1, tensor_2)

    if not merge_physicals:
        return product.reshape(
            (
                tensor_1.shape[0] * tensor_2.shape[0],
                tensor_1.shape[1],
                tensor_2.shape[1],
                tensor_1.shape[2] * tensor_2.shape[2],
            )
        )

    return product


def split_two_site_tensor(
    tensor: np.ndarray,
    chi_max: np.int16 = 1e4,
    cut: np.float64 = 1e-12,
    renormalise: bool = False,
) -> tuple:
    """
    Split a two-site MPS tensor according to the following diagram::

       i ---(tensor)--- l     ->    i ---(A)---diag(S)---(B)--- l
             |   |                         |               |
             j   k                         j               k

    Parameters
    ----------
    tensor : np.ndarray
        Two-site tensor ``(i, j, k, l)``.
    chi_max : np.int16
        Maximum number of singular values to keep.
    eps : np.float64
        Discard any singular values smaller than eps.

    Returns
    -------
    a_l : np.ndarray
        Left isometry ``(i, j, m)``.
    singular_values : np.ndarray
        List of singular values.
    b_r : np.ndarray
        Right isometry ``(m, k, l)``.

    Raises
    ------
    ValueError
        If the tensor is not four-dimensional.
    """

    if len(tensor.shape) != 4:
        raise ValueError(
            "A valid two-site tensor must have 4 legs"
            f"while the one given has {len(tensor.shape)}."
        )

    chi_v_l, d_l, d_r, chi_v_r = tensor.shape
    tensor = tensor.reshape((chi_v_l * d_l, d_r * chi_v_r))

    u_l, singular_values, v_r = svd(
        tensor, cut=cut, chi_max=chi_max, renormalise=renormalise
    )

    chi_v_cut = len(singular_values)
    a_l = u_l.reshape((chi_v_l, d_l, chi_v_cut))
    b_r = v_r.reshape((chi_v_cut, d_r, chi_v_r))

    return a_l, singular_values, b_r


def create_random_mpo(
    num_sites: np.int16,
    bond_dimensions: list[np.int16],
    phys_dim: np.int16,
    which: str = "uniform",
) -> list[np.ndarray]:
    """
    Creates a random complex-valued Matrix Product Operator.

    Parameters
    ----------
    num_sites : np.int16
        The number of sites for the MPO.
        This will be equal to the number of tensors.
    bond_dimensions : list[np.int16]
        A list of bond dimensions.
    phys_dim : np.int16
        Physical dimension of the tensors.
    which : str
        Specifies the distribution from which
        the matrix elements are being taken.
        Options: "uniform", "normal", "randint".

    Returns
    -------
    mpo : list[np.ndarray]
        The resulting MPO.

    Notes
    -----
    The ``bond_dimensions`` argument should be given as a list of right virtual dimensions
    without the last trivial virtual dimension. Thus, the length of the list is ``num_sites - 1``.

    The distributions available: uniform(0,1), normal, random integer {0,1}.

    Each tensor in the MPO list has legs ``(vL, vR, pU, pD)``, where ``v`` stands for "virtual",
    ``p`` -- for "physical", and ``L, R, U, D`` -- for "left", "right", "up", "down" accordingly.
    """

    bonds = [[dim, dim] for dim in bond_dimensions]
    bonds.append([1])
    bonds.insert(0, [1])
    bonds = [item for sublist in bonds for item in sublist]

    shapes = [
        (bonds[i], bonds[i + 1], phys_dim, phys_dim)
        for i in range(0, len(bonds) - 1, 2)
    ]

    if which == "uniform":
        mpo = [
            np.random.uniform(size=shapes[i]) + 1j * np.random.uniform(size=shapes[i])
            for i in range(num_sites)
        ]

    if which == "normal":
        mpo = [
            np.random.normal(size=shapes[i]) + 1j * np.random.normal(size=shapes[i])
            for i in range(num_sites)
        ]

    if which == "randint":
        mpo = [
            np.random.randint(0, 2, size=shapes[i])
            + 1j * np.random.randint(0, 2, size=shapes[i])
            for i in range(num_sites)
        ]

    return mpo


def mpo_to_matrix(
    mpo: list[np.ndarray], interlace: bool = False, group: bool = False
) -> np.ndarray:
    """
    Creates a matrix from an MPO.

    Parameters
    ----------
    mpo : list[np.ndarray]
        The MPO to convert to a matrix.
    interlace : bool
        Whether to interlace the matrix' legs or not.
    group : bool
        Whether to group the matrix' legs or not, see the notes.
        Grouping means merging all the up legs into one leg and the same for the down legs.

    Returns
    -------
    matrix : np.ndarray
        The resulting matrix.

    Raises
    ------
    ValueError
        If any of the MPO tensors is not four-dimensional.

    Notes
    -----
    If ``interlace==True``, the matrix' legs will go as
    ``(p0U, p0D, p1U, p1D, ...)``, which means
    physical legs sticking up and down with the site number.
    If ``interlace==False``, the matrix' legs will go as
    ``(p0D, p1D, ..., p0U, p1U, ...)``, which means listing first
    all physical legs sticking down with the site number,
    and then all physical legs sticking up.
    This is done to adjust the matrix to the ``@`` numpy-native matrix-vector multiplication.
    Note, grouping (if wanted) is being done after the interlacing.

    An example of a matrix with ungrouped legs on three sites::

          p0U p1U p2U
         __|___|___|__
        |            |
        |____________|
           |   |   |
          p0D p1D p2D

    Each tensor in the MPO list has legs ``(vL, vR, pU, pD)``, where v stands for "virtual",
    p -- for "physical", and L, R, U, D -- for "left", "right", "up", "down" accordingly.
    Warning: will cause memory overflow for number of sites > ~20.
    """

    for i, tensor in enumerate(mpo):
        if len(tensor.shape) != 4:
            raise ValueError(
                f"A valid MPO tensor must have 4 legs while tensor {i} has {len(tensor.shape)}."
            )

    phys_dim = mpo[0].shape[2]
    num_sites = len(mpo)
    matrix = (
        np.tensordot(mpo[0], mpo[1], (1, 0))
        .transpose((0, 3, 1, 2, 4, 5))
        .reshape((-1, mpo[1].shape[1], phys_dim**2, phys_dim**2))
    )

    for i in range(num_sites - 2):

        matrix = (
            np.tensordot(matrix, mpo[i + 2], (1, 0))
            .transpose((0, 3, 1, 2, 4, 5))
            .reshape(
                (-1, mpo[i + 2].shape[1], phys_dim ** (i + 3), phys_dim ** (i + 3))
            )
        )

    matrix = matrix.reshape(([phys_dim] * num_sites * 2))

    if not interlace:
        order = list(range(1, 2 * num_sites, 2)) + list(range(0, 2 * num_sites, 2))
        matrix = matrix.transpose(order)

    if group:
        matrix = matrix.reshape((phys_dim**num_sites, phys_dim**num_sites))

    return matrix


def mpo_from_matrix(
    matrix: np.ndarray,
    num_sites: np.int16,
    interlaced: bool = True,
    phys_dim: np.int16 = 2,
    chi_max: np.int16 = 1e4,
) -> list[np.ndarray]:
    """
    Creates an MPO from a matrix.

    Parameters
    ----------
    matrix : np.ndarray
        The matrix to convert to an MPO.
        Can be given with either physical legs grouped together or merged (see notes).
    num_sites : np.int16
        The number of sites in the MPO.
    interlaced : bool
        Whether the matrix legs are interlaced or not.
    phys_dim : np.int16
        Local dimension of the physical legs.
    chi_max : np.int16
        Maximum bond dimension allowed in the MPO.

    Returns
    -------
    mpo : list[np.ndarray]
        The resulting Matrix Product Operator.

    Raises
    ------
    ValueError
        If the matrix' shape does not correspond to ``phys_dim`` and ``num_sites``.

    Notes
    -----
    If ``interlaced==True``, the matrix' legs are considered to go as
    ``(p0U, p0D, p1U, p1D, ...)``, which means physical legs sticking up and down with the site number.
    If ``interlaced==False``, the matrix' legs are considered to go as
    ``(p0D, p1D, ..., p0U, p1U, ...)``, which means listing first all physical legs sticking down
    with the site number, and then all physical legs sticking up.
    This is done to adjust the matrix to the ``@`` numpy-native matrix-vector multiplication.

    An example of a matrix with ungrouped legs on three sites::

          p0U p1U p2U
         __|___|___|__
        |            |
        |____________|
           |   |   |
          p0D p1D p2D

    Each tensor in the ``mpo`` list will have legs ``(vL, vR, pU, pD)``, where ``v`` stands for "virtual",
    ``p`` -- for "physical", and ``L, R, U, D`` -- for "left", "right", "up", "down" accordingly.
    """

    hilbert_space_dim = phys_dim**num_sites
    phys_dims = [phys_dim] * num_sites * 2
    mps_dim = phys_dim**2

    # Checking the matrix has correct shapes for
    # a given physical dimension and number of sites.
    if (matrix.shape != tuple([hilbert_space_dim] * 2)) and (
        matrix.shape != tuple(phys_dims)
    ):
        raise ValueError(
            f"The matrix' shape should be either {tuple([hilbert_space_dim] * 2)}, "
            f"or {tuple(phys_dims)}, instead, the matrix given has shape {matrix.shape}."
        )

    # Copying the matrix not to change the original one inplace.
    matrix = matrix.copy()

    # Reshaping the matrix.
    if matrix.shape == tuple([hilbert_space_dim] * 2):
        matrix = matrix.reshape(tuple(phys_dims))

    # Dealing with possible interlacing.
    if not interlaced:
        correct_order = list([i + num_sites, i] for i in range(num_sites))
        correct_order = [item for sublist in correct_order for item in sublist]
        matrix = matrix.transpose(correct_order)

    # Treating the MPO as an MPS with squared physical dimensions.
    mpo = []
    matrix = matrix.reshape((-1, mps_dim))
    matrix, singular_values, v_r = svd(matrix, chi_max=chi_max, renormalise=False)
    v_r = np.matmul(np.diag(np.sqrt(singular_values)), v_r)
    mpo.append(np.expand_dims(v_r, -1))
    while matrix.shape[0] >= mps_dim:
        matrix = np.matmul(matrix, np.diag(np.sqrt(singular_values)))
        bond_dim = matrix.shape[-1]
        matrix = matrix.reshape((-1, mps_dim * bond_dim))
        matrix, singular_values, v_r = svd(matrix, chi_max=chi_max, renormalise=False)
        v_r = np.matmul(np.diag(np.sqrt(singular_values)), v_r)
        v_r = v_r.reshape((-1, mps_dim, bond_dim))
        mpo.insert(0, v_r)

    # Contracting in the orthogonality centre.
    mpo[0] = contract(
        "ij, jk, klm",
        matrix,
        np.diag(np.sqrt(singular_values)),
        mpo[0],
        optimize=[(0, 1), (0, 1)],
    )

    # Converting the MPS back to the MPO.
    mpo = [tensor.transpose((0, 2, 1)) for tensor in mpo]
    mpo = [
        tensor.reshape((tensor.shape[0], tensor.shape[1], phys_dim, phys_dim))
        for tensor in mpo
    ]

    return mpo



"""
This module contains the :class:`DMRG` and the :class:`EffectiveOperator` classes.
Inspired by TenPy.
"""

from typing import Union
import numpy as np
import scipy.sparse
from opt_einsum import contract
from scipy.sparse.linalg import eigsh
from tqdm import tqdm

from mdopt.mps.canonical import CanonicalMPS
from mdopt.mps.explicit import ExplicitMPS
from mdopt.utils.utils import split_two_site_tensor


class EffectiveOperator(scipy.sparse.linalg.LinearOperator):
    r"""Class to store an effective two-site operator.

    In order to take more advantage of the ``scipy.sparse.linalg`` module,
    we make a special class for local effective operators.
    It allows us to compute eigenvectors more effeciently.

    Such effective operator is to be diagonalised in the
    ``update_bond`` method of the :class:'DMRG' class::

        ---uL                    uR---
        |        i          j        |
        |  vL    |          |    vR  |
        (L)----(mpo_l)----(mpo_r)----(R)
        |        |          |        |
        |        k          l        |
        ---dL                    dR---

    In our convention, the legs of left/right environments (tensors ``L``/``R`` in the cartoon)
    are ordered as follows: ``(uL/uR, vL/vR, dL/dR)`` which means "(up, virtual, down)".
    """

    def __init__(
        self,
        left_environment: np.ndarray,
        mpo_tensor_left: np.ndarray,
        mpo_tensor_right: np.ndarray,
        right_environment: np.ndarray,
    ):
        """Initialises an effective operator tensor network.

        Parameters
        ----------
        left_environment : np.ndarray
            The left environment for the effective operator.
        mpo_tensor_left : np.ndarray
            The left MPO tensor.
        mpo_tensor_right : np.ndarray
            The right MPO tensor.
        right_environment : np.ndarray
            The right environment for the effective operator.
        """

        if len(left_environment.shape) != 3:
            raise ValueError(
                "A valid left environment tensor must have 3 legs"
                f"while the one given has {len(left_environment.shape)}."
            )
        if len(mpo_tensor_left.shape) != 4:
            raise ValueError(
                "A valid mpo left tensor must have 4 legs"
                f"while the one given has {len(mpo_tensor_left.shape)}."
            )
        if len(mpo_tensor_right.shape) != 4:
            raise ValueError(
                "A valid mpo right tensor must have 4 legs"
                f"while the one given has {len(mpo_tensor_right.shape)}."
            )
        if len(right_environment.shape) != 3:
            raise ValueError(
                "A valid right environment tensor must have 3 legs"
                f"while the one given has {len(right_environment.shape)}."
            )

        self.left_environment = left_environment
        self.right_environment = right_environment
        self.mpo_tensor_left = mpo_tensor_left
        self.mpo_tensor_right = mpo_tensor_right
        chi_1, chi_2 = (
            left_environment.shape[2],
            right_environment.shape[2],
        )
        d_1, d_2 = (
            mpo_tensor_left.shape[3],
            mpo_tensor_right.shape[3],
        )
        self.x_shape = (chi_1, d_1, d_2, chi_2)
        self.shape = (chi_1 * d_1 * d_2 * chi_2, chi_1 * d_1 * d_2 * chi_2)
        self.dtype = mpo_tensor_left.dtype
        super().__init__(shape=self.shape, dtype=self.dtype)

    def _matvec(self, x: np.ndarray) -> np.ndarray:
        """
        Performs matrix-vector multiplication.

        Computes ``effective_operator * |x> = |x'>``.
        This function is being used by :func:`scipy.sparse.linalg.eigsh` to diagonalise
        the effective operator with the Lanczos method, without generating the full matrix.

        Parameters
        ----------
        x : np.ndarray
            The two-site tensor on which acts an effective operator.
        """

        two_site_tensor = np.reshape(x, self.x_shape)

        if len(two_site_tensor.shape) != 4:
            raise ValueError(
                f"A valid two-site tensor must have 4 legs"
                f"while the one given has {len(two_site_tensor.shape)}."
            )

        einsum_string = "ijkl, mni, nopj, oqrk, sql -> mprs"
        two_site_tensor = contract(
            einsum_string,
            two_site_tensor,
            self.left_environment,
            self.mpo_tensor_left,
            self.mpo_tensor_right,
            self.right_environment,
            optimize=[(0, 1), (0, 3), (0, 2), (0, 1)],
        )

        return np.reshape(two_site_tensor, self.shape[0])


class DMRG:
    """
    Class storing the DMRG methods.

    Class holding the Density Matrix Renormalisation Group algorithm with two-site updates
    for a finite-size system with open-boundary conditions.

    Attributes
    ----------
    mps : Union[ExplicitMPS, CanonicalMPS]
        MPS serving as a current approximation of the target state.
    mpo : list[np.ndarray]
        The MPO of which the target state is to be computed.
        Each tensor in the MPO list has legs ``(vL, vR, pU, pD)``,
        where ``v`` stands for "virtual", ``p`` -- for "physical",
        and ``L``, ``R``, ``U``, ``D`` -- for "left", "right", "up", "down" accordingly.
    chi_max : np.int16
        The highest bond dimension of an MPS allowed.
    cut : np.float64
        The lower boundary of the spectrum, i.e., all the
        singular values smaller than that will be discarded.
    mode : str
        The eigensolver mode. Available options:
            | ``LM`` : Largest (in magnitude) eigenvalues.
            | ``SM`` : Smallest (in magnitude) eigenvalues.
            | ``LA`` : Largest (algebraic) eigenvalues.
            | ``SA`` : Smallest (algebraic) eigenvalues.
    silent : np.float64
        Whether to show/hide the progress bar.
    copy : bool
        Whether to copy the input MPS or modify inplace.
    """

    def __init__(
        self,
        mps: Union[ExplicitMPS, CanonicalMPS],
        mpo: list[np.ndarray],
        chi_max: np.int16 = 1e4,
        cut: np.float64 = 1e-12,
        mode: str = "SA",
        silent: bool = False,
        copy: bool = True,
    ):

        if len(mps) != len(mpo):
            raise ValueError(
                f"The MPS has length {len(mps)},"
                f"the MPO has length {len(mpo)},"
                "but the lengths should be equal."
            )
        for i, tensor in enumerate(mpo):
            if len(tensor.shape) != 4:
                raise ValueError(
                    f"A valid MPO tensor must have 4 legs while tensor {i} has {len(tensor.shape)}."
                )
        if mode not in ["SA", "LA", "SM", "LM"]:
            raise ValueError("Invalid eigensolver mode given.")

        self.mps = mps
        if copy:
            self.mps = mps.copy()
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.right_canonical()
        self.left_environments = [None] * len(mps)
        self.right_environments = [None] * len(mps)
        self.mpo = mpo
        self.chi_max = chi_max
        self.cut = cut
        self.mode = mode
        self.silent = silent

        start_bond_dim = self.mpo[0].shape[0]
        chi = mps.tensors[0].shape[0]
        left_environment = np.zeros([chi, start_bond_dim, chi], dtype=np.float64)
        right_environment = np.zeros([chi, start_bond_dim, chi], dtype=np.float64)
        left_environment[:, 0, :] = np.eye(chi, dtype=np.float64)
        right_environment[:, start_bond_dim - 1, :] = np.eye(chi, dtype=np.float64)
        self.left_environments[0] = left_environment
        self.right_environments[-1] = right_environment
        for i in reversed(range(1, len(mps))):
            self.update_right_environment(i)

    def sweep(self):
        """
        One DMRG sweep.

        A method performing one DMRG sweep, which consists of
        two series of ``update_bond`` sweeps which go back and forth.
        """

        for i in range(self.mps.num_sites - 1):
            self.update_bond(i)

        for i in reversed(range(self.mps.num_sites - 1)):
            self.update_bond(i)

    def update_bond(self, i: np.int16):
        """
        Updates the bond between sites ``i`` and ``i+1``.
        """

        j = i + 1

        effective_hamiltonian = EffectiveOperator(
            self.left_environments[i],
            self.mpo[i],
            self.mpo[j],
            self.right_environments[j],
        )

        if isinstance(self.mps, ExplicitMPS):
            initial_guess = self.mps.two_site_right_iso(i).reshape(
                effective_hamiltonian.shape[0]
            )
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i)
            initial_guess = self.mps.two_site_tensor_next(i).reshape(
                effective_hamiltonian.shape[0]
            )

        _, eigenvectors = eigsh(
            effective_hamiltonian,
            k=1,
            which=self.mode,
            return_eigenvectors=True,
            v0=initial_guess,
            tol=1e-8,
        )
        x = eigenvectors[:, 0].reshape(effective_hamiltonian.x_shape)
        left_iso_i, singular_values_j, right_iso_j = split_two_site_tensor(
            x, chi_max=self.chi_max, cut=self.cut, renormalise=True
        )

        if isinstance(self.mps, CanonicalMPS):
            self.mps.tensors[i] = np.tensordot(
                left_iso_i, np.diag(singular_values_j), (2, 0)
            )
            self.mps.orth_centre = i
            self.mps.tensors[j] = right_iso_j

        if isinstance(self.mps, ExplicitMPS):
            self.mps.tensors[i] = np.tensordot(
                np.linalg.inv(np.diag(self.mps.singular_values[i])), left_iso_i, (1, 0)
            )
            self.mps.tensors[j] = np.tensordot(
                right_iso_j,
                np.linalg.inv(np.diag(self.mps.singular_values[j + 1])),
                (2, 0),
            )
            self.mps.singular_values[j] = singular_values_j

        self.update_left_environment(i)
        self.update_right_environment(j)

    def update_right_environment(self, i: np.int16):
        """
        Compute the ``right_environment`` right of site ``i-1``
        from the ``right_environment`` right of site ``i``.
        """

        right_environment = self.right_environments[i]

        if isinstance(self.mps, ExplicitMPS):
            right_iso = self.mps.single_site_right_iso(i)
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i - 1)
            right_iso = self.mps.single_site_tensor(i)

        right_environment = contract(
            "ijk, lnjm, omp, knp -> ilo",
            right_iso,
            self.mpo[i],
            np.conjugate(right_iso),
            right_environment,
            optimize=[(0, 3), (0, 2), (0, 1)],
        )
        self.right_environments[i - 1] = right_environment

    def update_left_environment(self, i: np.int16):
        """
        Compute the ``left_environment`` left of site ``i+1``
        from the ``left_environment`` left of site ``i``.
        """

        left_environment = self.left_environments[i]

        if isinstance(self.mps, ExplicitMPS):
            left_iso = self.mps.single_site_left_iso(i)
        if isinstance(self.mps, CanonicalMPS):
            self.mps = self.mps.move_orth_centre(i + 1)
            left_iso = self.mps.single_site_tensor(i)

        left_environment = contract(
            "ijk, lnjm, omp, ilo -> knp",
            left_iso,
            self.mpo[i],
            np.conjugate(left_iso),
            left_environment,
            optimize=[(0, 3), (0, 2), (0, 1)],
        )
        self.left_environments[i + 1] = left_environment

    def run(self, num_iter: np.int16 = 1):
        """
        Run the algorithm, i.e., run the ``sweep`` method for ``num_iter`` number of iterations.
        """

        for _ in tqdm(range(num_iter), disable=self.silent):
            self.sweep()



"""
This module contains the :class:`CanonicalMPS` class.
Hereafter, by saying the Matrix Product State (MPS) is in a canonical form we mean one of the following.

    1) Right-canonical: all tensors are right isometries, i.e.::

        ---( )---     ---
            |   |       |
            |   |  ==   |
            |   |       |
        ---(*)---     ---

    2) Left-canonical: all tensors are left isometries, i.e.::

        ---( )---     ---
        |   |         |
        |   |    ==   |
        |   |         |
        ---(*)---     ---

    3) Mixed-canonical: all but one tensors are left or right isometries.
    This exceptional tensor will be hereafter called the **orthogonality centre**.

    Note that, in the diagrams, a tensor with a star inside means that it is complex-conjugated.

    The Matrix Product State is stored as a list of three-dimensional tensors.
    Essentially, it corresponds to storing each ``A[i]`` or ``B[i]`` as shown in
    fig.4c in reference `[1]`_.

    .. _[1]: https://arxiv.org/abs/1805.00055
"""

from functools import reduce
from copy import deepcopy
from typing import Optional, Iterable, Tuple
import numpy as np

import mdopt
from mdopt.utils.utils import kron_tensors, split_two_site_tensor


class CanonicalMPS:
    """
    Class for finite-size canonical matrix product states with open boundary conditions.

    Attributes
    ----------
    tensors : list[np.ndarray]
        The tensors of the MPS, one per each physical site.
        Each tensor has legs (virtual left, physical, virtual right), in short ``(vL, i, vR)``.
    orth_centre : Optional[np.int16]
        Position of the orthogonality centre, does not support negative indexing.
        As a convention, this attribute is taken ``0`` for a right-canonical form,
        ``len(tensors) - 1`` for a left-canonical form, ``None`` for a product state.
    tolerance : np.float64
        Numerical tolerance to zero out the singular values in Singular Value Decomposition.
    chi_max : np.int16
        The maximum bond dimension to keep in Singular Value Decompositions.
    bond_dimensions : list[np.int16]
        The list of all bond dimensions of the MPS.
    phys_dimensions : list[np.int16]
        The list of all physical dimensions of the MPS.
    num_sites : np.int16
        Number of sites.
    num_bonds : np.int16
        Number of bonds.

    Raises
    ------
    ValueError
        If the orthogonality centre position does not correspond to the number of sites.
    ValueError
        If any of the tensors does not have three dimensions.
    """

    def __init__(
        self,
        tensors: list[np.ndarray],
        orth_centre: Optional[np.int16] = None,
        tolerance: np.float64 = 1e-12,
        chi_max: np.int16 = 1e4,
    ):

        self.tensors = tensors
        self.num_sites = len(tensors)
        self.num_bonds = self.num_sites - 1
        self.bond_dimensions = [self.tensors[i].shape[2] for i in range(self.num_bonds)]
        self.phys_dimensions = [self.tensors[i].shape[1] for i in range(self.num_sites)]
        self.orth_centre = orth_centre
        self.dtype = tensors[0].dtype
        self.tolerance = tolerance
        self.chi_max = chi_max

        if orth_centre and orth_centre not in range(self.num_sites):
            raise ValueError(
                f"The orthogonality centre index must reside anywhere from site 0"
                f"to {self.num_sites-1}, the one given is at position {orth_centre}."
            )

        for _, tensor in enumerate(tensors):
            if len(tensor.shape) != 3:
                raise ValueError(
                    "A valid MPS tensor must have 3 legs"
                    f"while the one given has {len(tensor.shape)}."
                )

    def __len__(self) -> np.int16:
        """
        Returns the number of sites in the MPS.
        """
        return self.num_sites

    def copy(self) -> "CanonicalMPS":
        """
        Returns a copy of the current MPS.
        """
        return CanonicalMPS(
            deepcopy(self.tensors), self.orth_centre, self.tolerance, self.chi_max
        )

    def reverse(self) -> "CanonicalMPS":
        """
        Returns a reversed version of a given MPS.
        """

        reversed_tensors = [np.transpose(tensor) for tensor in reversed(self.tensors)]
        if self.orth_centre:
            reversed_orth_centre = (self.num_sites - 1) - self.orth_centre
            return CanonicalMPS(
                reversed_tensors, reversed_orth_centre, self.tolerance, self.chi_max
            )

        return CanonicalMPS(reversed_tensors, None, self.tolerance, self.chi_max)

    def conjugate(self) -> "CanonicalMPS":
        """
        Returns a complex-conjugated version of the current MPS.
        """

        conjugated_tensors = [np.conjugate(mps_tensor) for mps_tensor in self.tensors]
        return CanonicalMPS(
            conjugated_tensors, self.orth_centre, self.tolerance, self.chi_max
        )

    def single_site_tensor(self, site: int) -> np.ndarray:
        """
        Returs a particular MPS tensor located at the corresponding site.

        Parameters
        ----------
        site: int
            The site index of the tensor.
        """

        if site not in range(self.num_sites):
            raise ValueError(
                f"Site given {site}, with the number of sites in the MPS {self.num_sites}."
            )

        return self.tensors[site]

    def single_site_tensor_iter(self) -> Iterable:
        """
        Returns an iterator over the single-site tensors for every site.
        """
        return (self.single_site_tensor(i) for i in range(self.num_sites))

    def two_site_tensor_next(self, site: int) -> np.ndarray:
        """
        Computes a two-site tensor on a given site and the next one.

        Parameters
        ----------
        site : int
            The site index of the tensor.
        """

        if site not in range(self.num_sites - 1):
            raise ValueError(
                f"Sites given {site}, {site + 1}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            self.single_site_tensor(site),
            self.single_site_tensor(site + 1),
            (2, 0),
        )

    def two_site_tensor_prev(self, site: int) -> np.ndarray:
        """
        Computes a two-site tensor on a given site and the previous one.

        Parameters
        ----------
        site : int
            The site index of the tensor.
        """

        if site not in range(1, self.num_sites):
            raise ValueError(
                f"Sites given {site - 1}, {site}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            self.single_site_tensor(site - 1),
            self.single_site_tensor(site),
            (2, 0),
        )

    def two_site_tensor_next_iter(self) -> Iterable:
        """
        Returns an iterator over the two-site tensors for every site and its right neighbour.
        """
        return (self.two_site_tensor_next(i) for i in range(self.num_sites - 1))

    def two_site_tensor_prev_iter(self) -> Iterable:
        """
        Returns an iterator over the two-site tensors for every site and its left neighbour.
        """
        return (self.two_site_tensor_prev(i) for i in range(1, self.num_sites))

    def dense(self, flatten: bool = True) -> np.ndarray:
        """
        Returns a dense representation of an MPS, given as a list of tensors.

        Warning: this method will cause memory overload for number of sites > ~20!

        Parameters
        ----------
        flatten : bool
            Whether to merge all the physical indices to form a vector or not.
        """

        dense = reduce(lambda a, b: np.tensordot(a, b, (-1, 0)), self.tensors)

        if flatten:
            return dense.flatten()

        return dense

    def density_mpo(self) -> list[np.ndarray]:
        """
        Returns the MPO representation (as a list of tensors)
        of the density matrix defined by the MPS in a canonical form.
        Each tensor in the MPO list has legs ``(vL, vR, pU, pD)``,
        where ``v`` stands for "virtual", ``p`` -- for "physical",
        and ``L, R, U, D`` stand for "left", "right", "up", "down".

        Notes
        -----
        This operation is depicted in the following diagram::

                   i     j
             a     |     |    c                 i     j
            ...---(*)---(*)---...           ab  |     |  cd
                                    -->  ...---[ ]---[ ]---...
            ...---( )---( )---...               |     |
             b     |     |    d                 k     l
                   k     l

        In the cartoon, ``{i,j,k,l}`` and ``{a,b,c,d}`` are indices.
        Here, the ``( )``'s represent the MPS tensors, the ``[ ]``'s ---the MPO tensors.
        The MPS with the physical legs up is complex-conjugated element-wise.
        The empty line between the MPS and its complex-conjugated version
        stands in fact for the tensor (kronecker) product.

        Warning, this object can be memory-intensive for large bond dimensions!
        """

        mpo = map(
            lambda t: kron_tensors(
                t, t, conjugate_second=True, merge_physicals=False
            ).transpose((0, 3, 2, 1)),
            self.tensors,
        )

        return list(mpo)

    def entanglement_entropy(self) -> np.ndarray:
        """
        Returns the entanglement entropy for bipartitions at each of the bonds.
        """
        return self.explicit().entanglement_entropy()

    def move_orth_centre(
        self, final_pos: np.int16, return_singular_values: bool = False
    ) -> tuple["CanonicalMPS", list]:
        """
        Moves the orthogonality centre from its current position to ``final_pos``.

        Returns a new version of the current :class:`CanonicalMPS` instance with
        the orthogonality centre moved from `self.orth_centre` to `final_pos`,
        returns also the singular value tensors from every covered bond as well.

        Parameters
        ----------
        final_pos : np.int16
            Final position of the orthogonality centre.
        return_singular_values : bool
            Whether to return the singular values obtained at each involved bond.

        Raises
        ------
        ValueError
            If ``final_pos`` does not match the MPS length.
        """

        if final_pos not in range(self.num_sites):
            raise ValueError(
                "The final position of the orthogonality centre should be"
                f"from 0 to {self.num_sites-1}, given {final_pos}."
            )

        singular_values = []

        if self.orth_centre is None:
            _, flags_left, flags_right = mdopt.mps.utils.find_orth_centre(
                self, return_orth_flags=True
            )
            if flags_left in (
                [True] + [False] * (self.num_sites - 1),
                [False] * self.num_sites,
            ):
                if flags_right == [not flag for flag in flags_left]:
                    self.orth_centre = 0
            if flags_left in (
                [True] * (self.num_sites - 1) + [False],
                [True] * self.num_sites,
            ):
                if flags_right == [not flag for flag in flags_left]:
                    self.orth_centre = self.num_sites - 1
            if (flags_left == [True] * self.num_sites) and (
                flags_right == [True] * self.num_sites
            ):
                self.orth_centre = 0

        if self.orth_centre < final_pos:
            begin, final = self.orth_centre, final_pos
            mps = self.copy()
        elif self.orth_centre > final_pos:
            mps = self.reverse()
            begin = mps.orth_centre
            final = (self.num_sites - 1) - final_pos
        else:
            return self

        for i in range(begin, final):
            two_site_tensor = mps.two_site_tensor_next(i)
            u_l, singular_values_bond, v_r = split_two_site_tensor(
                two_site_tensor,
                chi_max=self.chi_max,
                renormalise=True,
            )
            singular_values.append(singular_values_bond)
            mps.tensors[i] = u_l
            mps.tensors[i + 1] = np.tensordot(
                np.diag(singular_values_bond), v_r, (1, 0)
            )

        if self.orth_centre > final_pos:
            mps = mps.reverse()
            singular_values = list(reversed(singular_values))

        if return_singular_values:
            return mps, singular_values

        return mps

    def move_orth_centre_to_border(self) -> Tuple["CanonicalMPS", str]:
        """
        Moves the orthogonality centre from its current position to the closest border.

        Returns a new version of the current :class:`CanonicalMPS` instance with
        the orthogonality centre moved to the closest (from the current position) border.
        """

        if self.orth_centre is None:
            _, flags_left, flags_right = mdopt.mps.utils.find_orth_centre(
                self, return_orth_flags=True
            )
            if flags_left in (
                [True] + [False] * (self.num_sites - 1),
                [False] * self.num_sites,
            ):
                if flags_right == [not flag for flag in flags_left]:
                    return self.copy(), "first"
            if flags_left in (
                [True] * (self.num_sites - 1) + [False],
                [True] * self.num_sites,
            ):
                if flags_right == [not flag for flag in flags_left]:
                    return self.copy(), "last"
            if all(flags_left) and all(flags_right):
                return self.copy(), "first"

        else:
            if self.orth_centre <= int(self.num_bonds / 2):
                mps = self.move_orth_centre(final_pos=0, return_singular_values=False)
                return mps, "first"

            mps = self.move_orth_centre(
                final_pos=self.num_sites - 1, return_singular_values=False
            )
        return mps, "last"

    def explicit(self) -> "mdopt.mps.explicit.ExplicitMPS":
        """
        Transforms a :class:`CanonicalMPS` instance into a :class:`ExplicitMPS` instance.
        Essentially, retrieves each ``Γ[i]`` and ``Λ[i]`` from ``A[i]`` or ``B[i]``.
        See fig.4b in `[1]`_ for reference.
        """

        (mps_canonical, border) = self.move_orth_centre_to_border()

        if border == "first":
            self.orth_centre = 0
            mps_canonical, singular_values = self.move_orth_centre(
                self.num_sites - 1, return_singular_values=True
            )
        else:
            self.orth_centre = self.num_sites - 1
            mps_canonical, singular_values = self.move_orth_centre(
                0, return_singular_values=True
            )

        singular_values.insert(0, np.array([1.0]))
        singular_values.append(np.array([1.0]))

        explicit_tensors = []
        for i in range(self.num_sites):
            explicit_tensors.append(
                np.tensordot(
                    mps_canonical.tensors[i],
                    np.linalg.inv(np.diag(singular_values[i + 1])),
                    (2, 0),
                )
            )

        return mdopt.mps.explicit.ExplicitMPS(explicit_tensors, singular_values)

    def right_canonical(self) -> "CanonicalMPS":
        """
        Returns the current MPS in the right-canonical form.
        See eq.19 in `[1]`_ for reference.
        """

        return self.move_orth_centre(0)

    def left_canonical(self) -> "CanonicalMPS":
        """
        Returns the current MPS in the left-canonical form.
        See eq.19 in `[1]`_ for reference.
        """

        return self.move_orth_centre(self.num_sites - 1)



"""
This module contains the :class:`ExplicitMPS` class.
Hereafter by saying the MPS is in an explicit form we mean that
the state is stored in the following format: for each three-dimensional tensor
at site `i`, there exists a singular values diagonal matrix at bond ``i``::

            i     i    i+1    i
    ...---[ ]---( )---[ ]---( )---...
                    |           |
                    |           |

For "ghost" bonds at indices ``0``, ``L-1`` (i.e., bonds of dimension 1),
the corresponding singular value tensors at the boundaries
would be the identities of the same dimension.
We index sites with ``i`` from ``0`` to ``L-1``, with bond ``i`` being left of site ``i``.
Essentially, it corresponds to storing each ``Γ[i]`` and ``Λ[i]`` as shown in
fig.4b in reference `[1]`_.
"""

from functools import reduce
from copy import deepcopy
from typing import Iterable
import numpy as np
from opt_einsum import contract

from mdopt.mps.canonical import CanonicalMPS
from mdopt.utils.utils import kron_tensors


class ExplicitMPS:
    """
    Class for finite-size explicit matrix product states (MPS) with open boundary conditions.

    Attributes
    ----------
    tensors : list[np.ndarray]
        The "physical" tensors of the MPS, one for each physical site.
        Each tensor has legs (virtual left, physical, virtual right), in short ``(vL, i, vR)``.
    singular_values : list[np.ndarray]
        The singular values at each of the bonds, ``singular_values[i]`` is left of ``tensors[i]``.
        Each singular values list at each bond is normalised to 1.
    tolerance : np.float64
        Absolute tolerance of the normalisation of the singular value spectrum at each bond.
    num_sites : np.int16
        Number of sites.
    num_bonds : np.int16
        Number of non-trivial bonds, which is equal to ``num_sites - 1``.

    Raises
    ------
    ValueError
        If the ``tensors`` and the ``singular_values`` lists do not have corresponding lengths.
        The number of singular value matrices should be equal to the number of tensors + 1,
        because there are two trivial singular value matrices at each of the ghost bonds.
    ValueError
        If any of the MPS tensors is not three-dimensional.
    ValueError
        If any of the singular-values tensors is not normalised within the `tolerance` attribute.
    """

    def __init__(
        self,
        tensors: list[np.ndarray],
        singular_values: list[np.ndarray],
        tolerance: np.float64 = 1e-12,
        chi_max: np.int16 = 1e4,
    ):

        self.tensors = tensors
        self.num_sites = len(tensors)
        self.num_bonds = self.num_sites - 1
        self.bond_dimensions = [self.tensors[i].shape[2] for i in range(self.num_bonds)]
        self.phys_dimensions = [self.tensors[i].shape[1] for i in range(self.num_sites)]
        self.singular_values = singular_values
        self.num_singval_mat = len(singular_values)
        self.dtype = tensors[0].dtype
        self.tolerance = tolerance
        self.chi_max = chi_max

        if self.num_sites != self.num_singval_mat - 1:
            raise ValueError(
                f"The number of tensors {self.num_sites} should correspond "
                "to the number of non-trivial singular value matrices "
                f"{len(tensors) - 1}, instead the number of "
                f"non-trivial singular value matrices is {self.num_singval_mat - 2}."
            )

        for i, tensor in enumerate(tensors):
            if len(tensor.shape) != 3:
                raise ValueError(
                    "A valid MPS tensor must have 3 legs"
                    f"while the one given has {len(tensor.shape)}."
                )

        for i, _ in enumerate(singular_values):
            norm = np.linalg.norm(singular_values[i])
            if abs(norm - 1) > tolerance:
                raise ValueError(
                    "The norm of each singular values tensor must be 1, "
                    f"instead the norm is {norm} at bond {i + 1}."
                )

    def __len__(self) -> np.int16:
        """Returns the number of sites in the MPS."""
        return self.num_sites

    def __iter__(self) -> Iterable:
        """Returns an iterator over (singular_values, tensors) pair for each site."""
        return zip(self.singular_values, self.tensors)

    def copy(self) -> "ExplicitMPS":
        """Returns a copy of the current MPS."""
        return ExplicitMPS(
            deepcopy(self.tensors),
            deepcopy(self.singular_values),
            self.tolerance,
            self.chi_max,
        )

    def reverse(self) -> "ExplicitMPS":
        """Returns a reversed version of a given MPS."""

        reversed_tensors = list(np.transpose(t) for t in reversed(self.tensors))
        reversed_singular_values = list(reversed(self.singular_values))

        return ExplicitMPS(
            reversed_tensors, reversed_singular_values, self.tolerance, self.chi_max
        )

    def conjugate(self) -> "ExplicitMPS":
        """Returns a complex-conjugated version of the current MPS."""

        conjugated_tensors = [np.conjugate(tensor) for tensor in self.tensors]
        conjugated_sing_vals = [
            np.conjugate(sing_vals) for sing_vals in self.singular_values
        ]
        return ExplicitMPS(
            conjugated_tensors,
            conjugated_sing_vals,
            self.tolerance,
            self.chi_max,
        )

    def single_site_left_iso(self, site: int) -> np.ndarray:
        """Computes a single-site left isometry at a given site."""

        if site not in range(self.num_sites):
            raise ValueError(
                f"Site given {site}, with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            np.diag(self.singular_values[site]), self.tensors[site], (1, 0)
        )

    def single_site_right_iso(self, site: int) -> np.ndarray:
        """Computes a single-site right isometry at a given site."""

        if site not in range(self.num_sites):
            raise ValueError(
                f"Sites given {site}, {site + 1}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            self.tensors[site], np.diag(self.singular_values[site + 1]), (2, 0)
        )

    def single_site_left_iso_iter(self) -> Iterable:
        """Returns an iterator over the left isometries for every site."""

        return (self.single_site_left_iso(i) for i in range(self.num_sites))

    def single_site_right_iso_iter(self) -> Iterable:
        """Returns an iterator over the right isometries for every site."""

        return (self.single_site_right_iso(i) for i in range(self.num_sites))

    def two_site_left_iso(self, site: int) -> np.ndarray:
        """Computes a two-site isometry on a given site and
        the following one from two single-site left isometries.
        """

        if site not in range(self.num_sites):
            raise ValueError(
                f"Sites given {site}, {site + 1}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            self.single_site_left_iso(site),
            self.single_site_left_iso(site + 1),
            (2, 0),
        )

    def two_site_right_iso(self, site: int) -> np.ndarray:
        """Computes a two-site isometry on a given site and
        the following one from two single-site right isometries.
        """

        if site not in range(self.num_sites):
            raise ValueError(
                f"Sites given {site}, {site + 1}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        return np.tensordot(
            self.single_site_right_iso(site),
            self.single_site_right_iso(site + 1),
            (2, 0),
        )

    def two_site_right_iso_iter(self) -> Iterable:
        """Returns an iterator over the two-site right isometries for every site and
        its right neighbour.
        """
        return (self.two_site_right_iso(i) for i in range(self.num_sites))

    def two_site_left_iso_iter(self) -> Iterable:
        """Returns an iterator over the two-site left isometries for every site and
        its right neighbour.
        """
        return (self.two_site_left_iso(i) for i in range(self.num_sites))

    def dense(self, flatten: bool = True) -> np.ndarray:
        """Returns dense representation of the MPS.

        Warning: will cause memory overload for number of sites > ~20!
        """

        tensors = list(self.single_site_right_iso_iter())
        dense = reduce(lambda a, b: np.tensordot(a, b, (-1, 0)), tensors)

        if flatten:
            return dense.flatten()

        return dense

    def density_mpo(self) -> list[np.ndarray]:
        """
        Returns the MPO representation (as a list of tensors)
        of the density matrix defined by a given MPS.
        This operation is depicted in the following picture::

                   i          j
              a    |          |        c           i     j
            ...---(*)---O*---(*)---O*---...    ab  |     |  cd
                                        --> ...---[ ]---[ ]---...
            ...---( )---O----( )---O----...        |     |
              b    |          |        d           k     l
                   k          l

        In the cartoon, `{i,j,k,l}` and `{a,b,c,d}` are single indices,
        while `ab` and `cd` denote multi indices.
        Here, the ( )'s represent the MPS tensors, the O's ---
        the singular values tensors, the [ ]'s --- the MPO tensors.
        The MPS with the physical legs up is complex-conjugated element-wise,
        this is denoted by the star sign.
        The empty line between the MPS and its complex-conjugated version
        stands for the tensor (kronecker) product.

        Each tensor in the MPO list has legs (vL, vR, pU, pD),
        where v stands for "virtual", p -- for "physical",
        and L, R, U, D stand for "left", "right", "up", "down".
        """

        tensors = list(self.single_site_right_iso_iter())

        mpo = map(
            lambda t: kron_tensors(
                t, t, conjugate_second=True, merge_physicals=False
            ).transpose((0, 3, 2, 1)),
            tensors,
        )

        return list(mpo)

    def entanglement_entropy(self) -> np.ndarray:
        """Returns the entanglement entropy for bipartitions at each of the bonds."""

        def xlogx(arg):
            if arg == 0:
                return 0
            return arg * np.log(arg)

        entropy = np.zeros(shape=(self.num_bonds,), dtype=np.float64)

        for bond in range(self.num_bonds):
            singular_values = self.singular_values[bond].copy()
            singular_values[singular_values < self.tolerance] = 0
            singular_values2 = singular_values * singular_values
            entropy[bond] = -np.sum(
                np.fromiter((xlogx(s) for s in singular_values2), dtype=np.float64)
            )
        return entropy

    def right_canonical(self) -> CanonicalMPS:
        """Returns the MPS in the right-canonical form given the MPS in the explicit form.

        (see eq.19 in https://scipost.org/10.21468/SciPostPhysLectNotes.5 for reference),
        """

        return CanonicalMPS(
            list(self.single_site_right_iso_iter()),
            orth_centre=None,
            tolerance=self.tolerance,
            chi_max=self.chi_max,
        )

    def left_canonical(self) -> CanonicalMPS:
        """Returns the MPS in the left-canonical form given the MPS in the explicit form.

        (see eq.19 in https://scipost.org/10.21468/SciPostPhysLectNotes.5 for reference),
        """

        return CanonicalMPS(
            list(self.single_site_left_iso_iter()),
            orth_centre=None,
            tolerance=self.tolerance,
            chi_max=self.chi_max,
        )

    def mixed_canonical(self, orth_centre: int) -> CanonicalMPS:
        """Returns the MPS in the mixed-canonical form
        with the orthogonality centre being located at `orth_centre`.

        Parameters
            orth_centre_index: int
                An integer which can take values `0, 1, ..., num_sites-1`.
                Denotes the position of the orthogonality centre --
                the only non-isometry in the new MPS.
        """

        if orth_centre not in range(self.num_sites):
            raise ValueError(
                f"Orthogonality centre index given {orth_centre}, "
                f"with the number of sites in the MPS {self.num_sites}."
            )

        mixed_can_mps = []

        for i in range(orth_centre):
            mixed_can_mps.append(self.single_site_left_iso(i))

        orth_centre_tensor = contract(
            "ij, jkl, lm -> ikm",
            np.diag(self.singular_values[orth_centre]),
            self.tensors[orth_centre],
            np.diag(self.singular_values[orth_centre + 1]),
            optimize=[(0, 1), (0, 1)],
        )
        mixed_can_mps.append(orth_centre_tensor)

        for i in range(orth_centre + 1, self.num_sites):
            mixed_can_mps.append(self.single_site_right_iso(i))

        return CanonicalMPS(
            mixed_can_mps,
            orth_centre=orth_centre,
            tolerance=self.tolerance,
            chi_max=self.chi_max,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mdopt', 'mdopt.contractor', 'mdopt.mps', 'mdopt.optimiser', 'mdopt.utils']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx>=5.2.3,<6.0.0',
 'matplotlib>=3.6.1,<4.0.0',
 'more-itertools>=8.12.0,<9.0.0',
 'numpy>=1.20.1,<2.0.0',
 'opt-einsum>=3.3.0,<4.0.0',
 'scipy>=1.9.2,<2.0.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0']

setup_kwargs = {
    'name': 'mdopt',
    'version': '0.2.0',
    'description': 'Discrete optimization in the tensor-network (specifically, MPS-MPO) language.',
    'long_description': 'None',
    'author': 'Aleksandr Berezutskii',
    'author_email': 'berezutskii@phystech.edu',
    'maintainer': 'Aleksandr Berezutskii',
    'maintainer_email': 'berezutskii@phystech.edu',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



"""
The ``mdopt.contractor.contractor`` module.
======================================================
This module contains the MPS-MPO contractor functions.
"""

from typing import Union
import numpy as np
from opt_einsum import contract

from mdopt.mps.canonical import CanonicalMPS
from mdopt.mps.explicit import ExplicitMPS
from mdopt.utils.utils import split_two_site_tensor


def apply_one_site_operator(tensor: np.ndarray, operator: np.ndarray) -> np.ndarray:
    """
    Applies a one-site operator to a canonical MPS as follows::

        ----(tensor)---  ->  ---(tensor_updated)---
               |                       |
           (operator)                  |
               |                       |

    The operator can be non-unitary, however note
    that a non-unitary operator might break the canonical form.
    The operator has legs ``(pU, pD)``, where ``p`` stands for "physical", and
    ``U``, ``D`` -- for "up", "down" accordingly.

    Parameters
    ----------
    tensor : np.ndarray
        The MPS tensor to apply the operator to.
    operator : np.ndarray
        The operator to be applied.

    Returns
    -------
    tensor_updated : np.ndarray
        The updated MPS tensor.

    Raises
    ------
    ValueError
        If the MPS tensor is not three-dimensional.
    ValueError
        If the operator tensor is not two-dimensional.

    """

    if len(tensor.shape) != 3:
        raise ValueError(
            f"A valid MPS tensor must have 3 legs while the one given has {len(tensor.shape)}."
        )

    if len(operator.shape) != 2:
        raise ValueError(
            "A valid single-site operator must have 2 legs"
            f"while the one given has {len(operator.shape)}."
        )

    tensor_updated = contract("ijk, jl -> ilk", tensor, operator, optimize=[(0, 1)])

    return tensor_updated


def apply_two_site_unitary(
    lambda_0: list,
    b_1: np.ndarray,
    b_2: np.ndarray,
    unitary: np.ndarray,
    chi_max: np.int16 = 1e4,
    cut: np.float64 = 1e-12,
) -> tuple[np.ndarray]:
    """
    Applies a two-site unitary operator to a right-canonical MPS as follows::


        ---(lambda_0)---(b_1)---(b_2)---  ->  ---(b_1_updated)---(b_2_updated)---
                          |      |                     |             |
                          (unitary)                    |             |
                          |      |                     |             |

    This function uses a trick which allows performing the contraction
    without computing the inverse of any singular value matrix,
    which can introduce numerical instabilities for small singular values.
    Returns back the resulting MPS tensors in the right-canonical form.

    Unitary has legs ``(pUL pUR, pDL pDR)``, where ``p`` stands for "physical", and
    ``L``, ``R``, ``U``, ``D`` -- for "left", "right", "up", "down" accordingly.

    Parameters
    ----------
    lambda_0 : list
        A list of singular values to the left of first MPS tensor.
    b_1 : np.ndarray
        The first MPS right-isometric tensor to apply the unitary to.
    b_2 : np.ndarray
        The second MPS right-isometric tensor to apply the unitary to.
    unitary : np.ndarray
        The unitary tensor we apply.
    cut: np.float64
        Singular values smaller than this will be discarded.
    chi_max: np.int16
        Maximum number of singular values to keep.

    Returns
    -------
    b_1_updated : np.ndarray
        The first updated MPS right-isometric tensor.
    b_2_updated : np.ndarray
        The second updated MPS right-isometric tensor.

    Raises
    ------
    ValueError
        If the first MPS tensor is not three-dimensional.
    ValueError
        If the second MPS tensor is not three-dimensional.
    ValueError
        If the operator tensor is not four-dimensional.

    """

    if len(b_1.shape) != 3:
        raise ValueError(
            f"A valid MPS tensor must have 3 legs while the b_1 given has {len(b_1.shape)}."
        )

    if len(b_2.shape) != 3:
        raise ValueError(
            f"A valid MPS tensor must have 3 legs while the b_2 given has {len(b_1.shape)}."
        )

    if len(unitary.shape) != 4:
        raise ValueError(
            "A valid two-site operator must have 4 legs"
            f"while the one given has {len(unitary.shape)}."
        )

    two_site_tensor_with_lambda_0 = contract(
        "ij, jkl, lmn -> ikmn", np.diag(lambda_0), b_1, b_2, optimize=[(0, 1), (0, 1)]
    )
    two_site_tensor_with_lambda_0 = contract(
        "ijkl, jkmn -> imnl", two_site_tensor_with_lambda_0, unitary, optimize=[(0, 1)]
    )

    two_site_tensor_wo_lambda_0 = contract("ijk, klm", b_1, b_2, optimize=[(0, 1)])
    two_site_tensor_wo_lambda_0 = contract(
        "ijkl, jkmn -> imnl", two_site_tensor_wo_lambda_0, unitary, optimize=[(0, 1)]
    )

    _, _, b_2_updated = split_two_site_tensor(
        two_site_tensor_with_lambda_0, chi_max=chi_max, cut=cut, renormalise=False
    )
    b_1_updated = contract(
        "ijkl, mkl -> ijm",
        two_site_tensor_wo_lambda_0,
        np.conjugate(b_2_updated),
        optimize=[(0, 1)],
    )

    return (b_1_updated, b_2_updated)


def mps_mpo_contract(
    mps: Union[ExplicitMPS, CanonicalMPS],
    mpo: list[np.ndarray],
    start_site: np.int16 = 0,
    renormalise: bool = False,
    chi_max: np.int64 = 1e4,
    cut: np.float64 = 1e-12,
    inplace: bool = False,
) -> CanonicalMPS:
    """
    Applies an MPO to an MPS.

    Applies an operator (not necessarily unitary) in the MPO format
    to a canonical MPS with the orthogonality centre at site ``start_site``
    while optionally renormalising singular values at each bond.
    Returning the updated MPS in the canonical form.

    Note, this algorithm goes from left to right.
    In order to run from right to left, reverse both the MPS and the MPO manually.

    The initial configuration looks as follows with
    ``---O---`` depicting the orthogonality centre::

        ...---( )---O----( )---( )---...---( )---...
                    |     |     |           |
                   [ ]---[ ]---[ ]---...---[ ]
                    |     |     |           |

    The contraction proceeds as described in the supplementary notes.

    Parameters
    ----------
    mps : Union[CanonicalMPS, ExplicitMPS]
        The initial MPS.
    mpo : list[np.ndarray]
        MPO as a list of tensors, where each tensor is corresponding to
        an operator applied at a certain site.
        The operators should be ordered in correspondence with the sites.
        According to our convention, each operator has legs (vL, vR, pU, pD),
        where v stands for "virtual", p -- for "physical",
        and L, R, U, D stand for "left", "right", "up", "down".
    start_site : np.int16
        Index of the starting site.
    renormalise : bool
        Whether to renormalise the singular values after each contraction
        involving two neigbouring MPS sites.
    chi_max : np.int16
        Maximum bond dimension to keep.
    cut : np.float64
        Cutoff for the singular values.
    inplace : bool
        Whether to modify the starting MPS or create a new one.

    Returns
    -------
    mps : CanonicalMPS
        The updated MPS in the canonical form.

    Raises
    ------
    ValueError
        If any of the MPO tensors is not four-dimensional.
    ValueError
        If the length of the MPO and the starting site doesn't correspond
        to the number of sites of the MPS.

    """

    if not inplace:
        mps = mps.copy()
    if isinstance(mps, ExplicitMPS):
        mps = mps.mixed_canonical(start_site)
    if mps.orth_centre != start_site:
        mps = mps.move_orth_centre(start_site)

    for i, tensor in enumerate(mpo):
        if len(tensor.shape) != 4:
            raise ValueError(
                f"A valid MPO tensor must have 4 legs while tensor {i} has {len(tensor.shape)}."
            )

    if start_site + len(mpo) > len(mps):
        raise ValueError(
            "The length of the MPO should correspond to the number of sites where it is applied, "
            f"given MPO of length {len(mpo)} and the starting site {start_site}, "
            f"while the MPS has length {len(mps)}."
        )

    orth_centre_index = start_site

    two_site_mps_mpo_tensor = contract(
        "ijk, klm, nojp, oqlr -> iprqm",
        mps.tensors[start_site],
        mps.tensors[start_site + 1],
        mpo[0],
        mpo[1],
        optimize=[(0, 1), (1, 2), (0, 1)],
    ).reshape(
        (
            mps.tensors[start_site].shape[0],
            mpo[0].shape[3],
            mpo[1].shape[3],
            mps.tensors[start_site + 1].shape[2] * mpo[1].shape[1],
        )
    )

    for i in range(len(mpo) - 2):

        mps.tensors[orth_centre_index], singular_values, b_r = split_two_site_tensor(
            two_site_mps_mpo_tensor, chi_max=chi_max, cut=cut, renormalise=renormalise
        )

        orth_centre_index += 1
        mps.orth_centre = orth_centre_index

        mps.tensors[orth_centre_index] = contract(
            "ij, jkl -> ikl", np.diag(singular_values), b_r, optimize=[(0, 1)]
        ).reshape(
            (
                len(singular_values),
                mpo[i + 1].shape[3],
                mpo[i + 1].shape[1],
                mps.tensors[orth_centre_index + 1].shape[0],
            )
        )

        two_site_mps_mpo_tensor = contract(
            "ijkl, lmn, komp -> ijpon",
            mps.tensors[orth_centre_index],
            mps.tensors[orth_centre_index + 1],
            mpo[i + 2],
            optimize=[(0, 1), (0, 1)],
        ).reshape(
            (
                len(singular_values),
                mps.tensors[orth_centre_index].shape[1],
                mpo[i + 2].shape[3],
                mps.tensors[orth_centre_index + 1].shape[2] * mpo[i + 2].shape[1],
            )
        )

    mps.tensors[orth_centre_index], singular_values, b_r = split_two_site_tensor(
        two_site_mps_mpo_tensor, chi_max=chi_max, cut=cut, renormalise=renormalise
    )

    mps.tensors[orth_centre_index + 1] = contract(
        "ij, jkl -> ikl", np.diag(singular_values), b_r, optimize=[(0, 1)]
    )

    return mps
